/*
 * paragraph.c
 * Copyright (C) 1998,1999 A.J. van Os
 *
 * Description:
 * Read the paragraph information from a MS Word file
 */

#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "antiword.h"

typedef enum row_info_tag {
	found_nothing,
	found_a_cell,
	found_end_of_row
} row_info_enum;


/*
 * iGet6InfoLength - the length of the information for Word 6/7 files
 */
static int
iGet6InfoLength(int iByteNbr, const unsigned char *aucFpage)
{
	switch (ucGetByte(iByteNbr, aucFpage)) {
	case 0x02: case 0x10: case 0x11: case 0x13:
	case 0x15: case 0x16: case 0x1b: case 0x1c:
	case 0x50: case 0x5d: case 0x60: case 0x61:
	case 0x63: case 0x65: case 0x6b: case 0x90:
	case 0xa4: case 0xa5: case 0xb6: case 0xb8:
	case 0xbd: case 0xc0:
		return 3;
	case 0x0c: case 0xbe: case 0xbf:
		return 2 + (int)ucGetByte(iByteNbr + 1, aucFpage);
	case 0x0f:
		return 3 + (int)usGetWord(iByteNbr + 1, aucFpage);
	case 0x14: case 0x4a:
		return 5;
	case 0x44:
		return 6;
	case 0x88: case 0x89:
		return 4;
	case 0xbb:
		return 13;
	default:
		return 2;
	}
} /* end of iGet6InfoLength */

/*
 * Returns 2 and fills the row information block if end of a row was found
 *	   1 if a cell was found
 *	   0 nothing useful was found
 */
static row_info_enum
eGet6RowInfo(int iFodo, const unsigned char *aucFpage, row_block_type *pRow)
{
	int	iBytes, iFodoOff, iInfoLen;
	int	iIndex, iSize, iCol;
	int	iPosCurr, iPosPrev;
	BOOL	bFound18, bFound19, bFoundBE;

	fail(iFodo <= 0 || aucFpage == NULL || pRow == NULL);

	iBytes = 2 * (int)ucGetByte(iFodo, aucFpage);
	iFodoOff = 3;
	bFound18 = FALSE;
	bFound19 = FALSE;
	bFoundBE = FALSE;
	while (iBytes >= iFodoOff + 1) {
		iInfoLen = 0;
		switch (ucGetByte(iFodo + iFodoOff, aucFpage)) {
		case 0x0f:
			if (iBytes < iFodoOff + 2) {
				iInfoLen = 2;
			}
			break;
		case 0x18:
			if (ucGetByte(iFodo + iFodoOff + 1,
					aucFpage) == 0x01) {
				bFound18 = TRUE;
			}
			break;
		case 0x19:
			if (ucGetByte(iFodo + iFodoOff + 1,
					aucFpage) == 0x01) {
				bFound19 = TRUE;
			}
			break;
		case 0xbe:
			iSize = (int)ucGetByte(iFodo + iFodoOff + 1, aucFpage);
			if (iSize < 6 || iBytes < iFodoOff + 7) {
				DBG_DEC(iSize);
				DBG_DEC(iFodoOff);
				iInfoLen = 2;
				break;
			}
			iCol = (int)ucGetByte(iFodo + iFodoOff + 3, aucFpage);
			if (iCol < 1 ||
			    iBytes < iFodoOff + 3 + (iCol + 1) * 2) {
				DBG_DEC(iCol);
				DBG_DEC(iFodoOff);
				iInfoLen = 2;
				break;
			}
			if (iCol >= elementsof(pRow->asColumnWidth)) {
				werr(1, "The number of columns is corrupt");
			}
			pRow->ucNumberOfColumns = iCol;
			pRow->iColumnWidthSum = 0;
			iPosPrev = (int)(short)usGetWord(
					iFodo + iFodoOff + 4,
					aucFpage);
			for (iIndex = 0; iIndex < iCol; iIndex++) {
				iPosCurr = (int)(short)usGetWord(
					iFodo + iFodoOff + 6 + iIndex * 2,
					aucFpage);
				pRow->asColumnWidth[iIndex] =
							iPosCurr - iPosPrev;
				pRow->iColumnWidthSum +=
					pRow->asColumnWidth[iIndex];
				iPosPrev = iPosCurr;
			}
			bFoundBE = TRUE;
			break;
		default:
			break;
		}
		if (iInfoLen <= 0) {
			iInfoLen =
				iGet6InfoLength(iFodo + iFodoOff, aucFpage);
			fail(iInfoLen <= 0);
		}
		iFodoOff += iInfoLen;
	}
	if (bFound18 && bFound19 && bFoundBE) {
		return found_end_of_row;
	}
	if (bFound18) {
		return found_a_cell;
	}
	return found_nothing;
} /* end of eGet6RowInfo */

/*
 * Fill the style information block with information
 * from a Word 6/7 file.
 * Returns TRUE when successful, otherwise FALSE
 */
static BOOL
bGet6StyleInfo(int iFodo, const unsigned char *aucFpage,
		style_block_type *pStyle)
{
	int	iBytes, iFodoOff, iInfoLen;
	int	iTmp;

	fail(iFodo <= 0 || aucFpage == NULL || pStyle == NULL);

	iBytes = 2 * (int)ucGetByte(iFodo, aucFpage);
	iFodoOff = 3;
	if (iBytes < 1) {
		return FALSE;
	}
	(void)memset(pStyle, 0, sizeof(*pStyle));
	pStyle->ucStyle = ucGetByte(iFodo + 1, aucFpage);
	while (iBytes >= iFodoOff + 1) {
		iInfoLen = 0;
		switch (ucGetByte(iFodo + iFodoOff, aucFpage)) {
		case 0x05:
			pStyle->ucAlignment = ucGetByte(
					iFodo + iFodoOff + 1, aucFpage);
			break;
		case 0x0c:
			iTmp = (int)ucGetByte(
					iFodo + iFodoOff + 1, aucFpage);
			if (iTmp >= 1) {
				pStyle->ucListType = ucGetByte(
					iFodo + iFodoOff + 2, aucFpage);
				pStyle->bInList = TRUE;
			}
			if (iTmp >= 21) {
				pStyle->ucListCharacter = ucGetByte(
					iFodo + iFodoOff + 22, aucFpage);
			}
			break;
		case 0x0d:
			iTmp = (int)ucGetByte(
					iFodo + iFodoOff + 1, aucFpage);
			switch (iTmp) {
			case 0x0c:
				pStyle->bUnmarked = TRUE;
				/* No break or return */
			case 0x0a:
			case 0x0b:
				pStyle->bInList = TRUE;
				break;
			default:
				DBG_HEX(iTmp);
				break;
			}
			break;
		case 0x0f:
			if (iBytes < iFodoOff + 2) {
				iInfoLen = 2;
			}
			break;
		case 0x10:
			pStyle->sRightIndent = (short)usGetWord(
					iFodo + iFodoOff + 1, aucFpage);
			break;
		case 0x11:
			pStyle->sLeftIndent = (short)usGetWord(
					iFodo + iFodoOff + 1, aucFpage);
			break;
		default:
			break;
		}
		if (iInfoLen <= 0) {
			iInfoLen =
				iGet6InfoLength(iFodo + iFodoOff, aucFpage);
			fail(iInfoLen <= 0);
		}
		iFodoOff += iInfoLen;
	}
	return TRUE;
} /* end of bGet6StyleInfo */

/*
 * Build the lists with Paragraph Information for Word 6/7 files
 */
static void
vGet6ParInfo(FILE *pFile, int iStartBlock,
	const int *aiBBD, int iBBDLen,
	const unsigned char *aucHeader)
{
	row_block_type		tRow;
	style_block_type	tStyle;
	unsigned short	*ausParfPage;
	unsigned char	*aucBuffer;
	size_t	tSize;
	int	iLenOld, iLen;
	int	iParmOffset, iParmOffsetFirst, iParmOffsetLast;
	int	iIndex, iIndex2, iCount, iOffset, iNbr, iFodo;
	int	iParfFirstPage, iBeginParfInfo,iParfInfoLen, iParfPageNum;
	unsigned char	aucFpage[BIG_BLOCK_SIZE];

	fail(pFile == NULL || aucHeader == NULL);
	fail(iStartBlock < 0);
	fail(aiBBD == NULL || iBBDLen < 0);

	iBeginParfInfo = (int)ulGetLong(0xc0, aucHeader);
	DBG_HEX(iBeginParfInfo);
	iParfInfoLen = (int)ulGetLong(0xc4, aucHeader);
	DBG_DEC(iParfInfoLen);

	aucBuffer = xmalloc(iParfInfoLen);
	if (!bReadBuffer(pFile, iStartBlock,
			aiBBD, iBBDLen, BIG_BLOCK_SIZE,
			aucBuffer, iBeginParfInfo, iParfInfoLen)) {
		xfree(aucBuffer);
		return;
	}
	NO_DBG_PRINT_BLOCK(aucBuffer, iParfInfoLen);

	iLen = (iParfInfoLen - 4) / 6;
	tSize = iLen * sizeof(unsigned short);
	ausParfPage = xmalloc(tSize);
	for (iIndex = 0, iOffset = (iLen + 1) * 4;
	     iIndex < iLen;
	     iIndex++, iOffset += 2) {
		 ausParfPage[iIndex] = usGetWord(iOffset, aucBuffer);
		 NO_DBG_DEC(ausParfPage[iIndex]);
	}
	DBG_HEX(ulGetLong(0, aucBuffer));
	xfree(aucBuffer);
	aucBuffer = NULL;
	iParfPageNum = (int)usGetWord(0x190, aucHeader);
	DBG_DEC(iParfPageNum);
	if (iLen < iParfPageNum) {
		/* Replace ParfPage by a longer version */
		iLenOld = iLen;
		iParfFirstPage = (int)usGetWord(0x18c, aucHeader);
		DBG_DEC(iParfFirstPage);
		iLen += iParfPageNum - 1;
		tSize = iLen * sizeof(unsigned short);
		ausParfPage = xrealloc(ausParfPage, tSize);
		/* Add new values */
		iCount = iParfFirstPage + 1;
		for (iIndex = iLenOld; iIndex < iLen; iIndex++) {
			ausParfPage[iIndex] = iCount;
			DBG_DEC(ausParfPage[iIndex]);
			iCount++;
		}
	}

	(void)memset(&tStyle, 0, sizeof(tStyle));
	(void)memset(&tRow, 0, sizeof(tRow));
	iParmOffsetFirst = -1;
	for (iIndex = 0; iIndex < iLen; iIndex++) {
		if (!bReadBuffer(pFile, iStartBlock,
				aiBBD, iBBDLen, BIG_BLOCK_SIZE,
				aucFpage,
				(int)ausParfPage[iIndex] * BIG_BLOCK_SIZE,
				BIG_BLOCK_SIZE)) {
			break;
		}
		iNbr = (int)ucGetByte(0x1ff, aucFpage);
		NO_DBG_DEC(iNbr);
		for (iIndex2 = 0; iIndex2 < iNbr; iIndex2++) {
			NO_DBG_HEX(ulGetLong(iIndex2 * 4, aucFpage));
			iFodo = 2 * (int)ucGetByte(
				(iNbr + 1) * 4 + iIndex2 * 7, aucFpage);
			if (iFodo <= 0) {
				continue;
			}
			if (bGet6StyleInfo(iFodo, aucFpage, &tStyle)) {
				iParmOffset = (int)ulGetLong(
						iIndex2 * 4, aucFpage);
				NO_DBG_HEX(iParmOffset);
				tStyle.iOffset = iTextOffset2FileOffset(
							iParmOffset);
				vAdd2StyleInfoList(&tStyle);
				(void)memset(&tStyle, 0, sizeof(tStyle));
			}
			switch (eGet6RowInfo(iFodo, aucFpage, &tRow)) {
			case found_a_cell:
				if (iParmOffsetFirst >= 0) {
					break;
				}
				iParmOffsetFirst = (int)ulGetLong(
						iIndex2 * 4, aucFpage);
				NO_DBG_HEX(iParmOffsetFirst);
				tRow.iOffsetStart = iTextOffset2FileOffset(
							iParmOffsetFirst);
				NO_DBG_HEX(tRow.iOffsetStart);
				break;
			case found_end_of_row:
				iParmOffsetLast = (int)ulGetLong(
						iIndex2 * 4, aucFpage);
				NO_DBG_HEX(iParmOffsetLast);
				tRow.iOffsetEnd = iTextOffset2FileOffset(
							iParmOffsetLast);
				NO_DBG_HEX(tRow.iOffsetEnd);
				vAdd2RowInfoList(&tRow);
				(void)memset(&tRow, 0, sizeof(tRow));
				iParmOffsetFirst = -1;
				break;
			default:
				break;
			}
		}
	}
	xfree(ausParfPage);
} /* end of vGet6ParInfo */

/*
 * Fill the font information block with information
 * from a Word 6/7 file.
 * Returns TRUE when successful, otherwise FALSE
 */
static BOOL
bGet6FontInfo(int iFodo, const unsigned char *aucFpage,
		font_block_type *pFont)
{
	int	iBytes, iFodoOff, iInfoLen;
	int	iTmp;

	fail(iFodo <= 0 || aucFpage == NULL || pFont == NULL);

	iBytes = (int)ucGetByte(iFodo, aucFpage);
	iFodoOff = 1;
	if (iBytes < 1) {
		return FALSE;
	}
	(void)memset(pFont, 0, sizeof(*pFont));
	pFont->ucFontsize = DEFAULT_FONT_SIZE;
	while (iBytes >= iFodoOff + 1) {
		iInfoLen = 0;
		switch (ucGetByte(iFodo + iFodoOff, aucFpage)) {
		case 0x0f:
			if (iBytes < iFodoOff + 2) {
				iInfoLen = 2;
			}
			break;
		case 0x55:
			iTmp = (int)ucGetByte(
					iFodo + iFodoOff + 1, aucFpage);
			if (iTmp == 0x01 || iTmp == 0x81) {
				pFont->ucFontstyle |= FONT_BOLD;
			}
			DBG_HEX_C(iTmp != 0x01 && iTmp != 0x81, iTmp);
			break;
		case 0x56:
			iTmp = (int)ucGetByte(
					iFodo + iFodoOff + 1, aucFpage);
			if (iTmp == 0x01 || iTmp == 0x81) {
				pFont->ucFontstyle |= FONT_ITALIC;
			}
			DBG_HEX_C(iTmp != 0x01 && iTmp != 0x81, iTmp);
			break;
		case 0x5a:
		case 0x5b:
			iTmp = (int)ucGetByte(
					iFodo + iFodoOff + 1, aucFpage);
			if (iTmp == 0x01 || iTmp == 0x81) {
				DBG_MSG("All capitals text");
				pFont->ucFontstyle |= FONT_CAPITALS;
			}
			DBG_HEX_C(iTmp != 0x01 && iTmp != 0x81, iTmp);
			break;
		case 0x5c:
			iTmp = (int)ucGetByte(
					iFodo + iFodoOff + 1, aucFpage);
			if (iTmp == 0x01 || iTmp == 0x81) {
				DBG_MSG("Hidden text");
				pFont->ucFontstyle |= FONT_HIDDEN;
			}
			DBG_HEX_C(iTmp != 0x01 && iTmp != 0x81, iTmp);
			break;
		case 0x5d:
			iTmp = (int)usGetWord(
					iFodo + iFodoOff + 1, aucFpage);
			if (iTmp >= 0 && iTmp <= UCHAR_MAX) {
				pFont->ucFontnumber = (unsigned char)iTmp;
			}
			break;
		case 0x62:
			pFont->ucFontcolour =
				ucGetByte(iFodo + iFodoOff + 1, aucFpage);
			break;
		case 0x63:
			iTmp = (int)usGetWord(
					iFodo + iFodoOff + 1, aucFpage);
			if (iTmp < 0) {
				pFont->ucFontsize = 0;
			} else if (iTmp > UCHAR_MAX) {
				pFont->ucFontsize = UCHAR_MAX;
			} else {
				pFont->ucFontsize = (unsigned char)iTmp;
			}
			break;
		default:
			break;
		}
		if (iInfoLen <= 0) {
			iInfoLen =
				iGet6InfoLength(iFodo + iFodoOff, aucFpage);
			fail(iInfoLen <= 0);
		}
		iFodoOff += iInfoLen;
	}
	return TRUE;
} /* end of bGet6FontInfo */

/*
 * Build the lists with Character Information for Word 6/7 files
 */
static void
vGet6ChrInfo(FILE *pFile, int iStartBlock,
	const int *aiBBD, int iBBDLen,
	const unsigned char *aucHeader)
{
	font_block_type		tFont;
	unsigned short	*ausCharPage;
	unsigned char	*aucBuffer;
	size_t	tSize;
	int	iLenOld, iLen;
	int	iParmOffset, iFileOffset;
	int	iIndex, iIndex2, iCount, iOffset, iNbr, iFodo;
	int	iCharFirstPage, iBeginCharInfo, iCharInfoLen, iCharPageNum;
	unsigned char	aucFpage[BIG_BLOCK_SIZE];

	fail(pFile == NULL || aucHeader == NULL);
	fail(iStartBlock < 0);
	fail(aiBBD == NULL || iBBDLen < 0);

	iBeginCharInfo = (int)ulGetLong(0xb8, aucHeader);
	DBG_HEX(iBeginCharInfo);
	iCharInfoLen = (int)ulGetLong(0xbc, aucHeader);
	DBG_DEC(iCharInfoLen);

	aucBuffer = xmalloc(iCharInfoLen);
	if (!bReadBuffer(pFile, iStartBlock,
			aiBBD, iBBDLen,BIG_BLOCK_SIZE,
			aucBuffer, iBeginCharInfo, iCharInfoLen)) {
		xfree(aucBuffer);
		return;
	}

	iLen = (iCharInfoLen - 4) / 6;
	tSize = iLen * sizeof(unsigned short);
	ausCharPage = xmalloc(tSize);
	for (iIndex = 0, iOffset = (iLen + 1) * 4;
	     iIndex < iLen;
	     iIndex++, iOffset += 2) {
		 ausCharPage[iIndex] = usGetWord(iOffset, aucBuffer);
		 NO_DBG_DEC(ausCharPage[iIndex]);
	}
	DBG_HEX(ulGetLong(0, aucBuffer));
	xfree(aucBuffer);
	aucBuffer = NULL;
#if defined(__hpux)
	/* Strange: the HPUX cpp can not deal with the number 0x18e */
	iCharPageNum = (int)aucHeader[0x18f] << 8 | (int)aucHeader[0x18e];
#else
	iCharPageNum = (int)usGetWord(0x18e, aucHeader);
#endif /* __hpux */
	DBG_DEC(iCharPageNum);
	if (iLen < iCharPageNum) {
		/* Replace CharPage by a longer version */
		iLenOld = iLen;
		iCharFirstPage = (int)usGetWord(0x18a, aucHeader);
		DBG_DEC(iCharFirstPage);
		iLen += iCharPageNum - 1;
		tSize = iLen * sizeof(unsigned short);
		ausCharPage = xrealloc(ausCharPage, tSize);
		/* Add new values */
		iCount = iCharFirstPage + 1;
		for (iIndex = iLenOld; iIndex < iLen; iIndex++) {
			ausCharPage[iIndex] = iCount;
			DBG_DEC(ausCharPage[iIndex]);
			iCount++;
		}
	}

	(void)memset(&tFont, 0, sizeof(tFont));
	for (iIndex = 0; iIndex < iLen; iIndex++) {
		if (!bReadBuffer(pFile, iStartBlock,
				aiBBD, iBBDLen, BIG_BLOCK_SIZE,
				aucFpage,
				(int)ausCharPage[iIndex] * BIG_BLOCK_SIZE,
				BIG_BLOCK_SIZE)) {
			break;
		}
		iNbr = (int)ucGetByte(0x1ff, aucFpage);
		NO_DBG_DEC(iNbr);
		for (iIndex2 = 0; iIndex2 < iNbr; iIndex2++) {
		  	iParmOffset = (int)ulGetLong(
						iIndex2 * 4, aucFpage);
			iFileOffset = iTextOffset2FileOffset(iParmOffset);
			iFodo = 2 * (int)ucGetByte(
				(iNbr + 1) * 4 + iIndex2, aucFpage);
			if (iFodo == 0) {
				NO_DBG_HEX(iParmOffset);
				vReset2FontInfoList(iFileOffset);
			} else {
				if (bGet6FontInfo(iFodo, aucFpage, &tFont)) {
					tFont.iOffset = iFileOffset;
					vAdd2FontInfoList(&tFont);
				}
			}
			(void)memset(&tFont, 0, sizeof(tFont));
		}
	}
	xfree(ausCharPage);
} /* end of vGet6ChrInfo */

/*
 * iGet8InfoLength - the length of the information for Word 8/97 files
 */
static int
iGet8InfoLength(int iByteNbr, const unsigned char *aucFpage)
{
	switch (usGetWord(iByteNbr, aucFpage)) {
	case 0x0835: case 0x0836: case 0x083a: case 0x083b: case 0x083c:
	case 0x2416: case 0x2417: case 0x260a:
		return 3;
	case 0x4600: case 0x460b: case 0x486d: case 0x4a43: case 0x4a4f:
	case 0x4a50: case 0x4a51: case 0x840f:
		return 4;
	case 0xd608:
		return 3 + (int)ucGetByte(iByteNbr + 2, aucFpage);
	default:
		return 1;
	}
} /* end of iGet8InfoLength */

/*
 * Returns 2 and fills the row information block if end of a row was found
 *	   1 if a cell was found
 *	   0 nothing useful was found
 */
static row_info_enum
eGet8RowInfo(int iFodo, const unsigned char *aucFpage, row_block_type *pRow)
{
	int	iBytes, iFodoOff, iInfoLen;
	int	iIndex, iSize, iCol;
	int	iPosCurr, iPosPrev;
	BOOL	bFound2416, bFound2417, bFoundd608;

	fail(iFodo <= 0 || aucFpage == NULL || pRow == NULL);

	iBytes = 2 * (int)ucGetByte(iFodo, aucFpage);
	iFodoOff = 3;
	if (iBytes == 0) {
		iFodo++;
		iBytes = 2 * (int)ucGetByte(iFodo, aucFpage);
	}
	NO_DBG_DEC(iBytes);
	bFound2416 = FALSE;
	bFound2417 = FALSE;
	bFoundd608 = FALSE;
	while (iBytes >= iFodoOff + 2) {
		iInfoLen = 0;
		switch (usGetWord(iFodo + iFodoOff, aucFpage)) {
		case 0x2416:
			if (ucGetByte(iFodo + iFodoOff + 2,
					aucFpage) == 0x01) {
				bFound2416 = TRUE;
			}
			break;
		case 0x2417:
			if (ucGetByte(iFodo + iFodoOff + 2,
					aucFpage) == 0x01) {
				bFound2417 = TRUE;
			}
			break;
		case 0xd608:
			iSize = (int)ucGetByte(iFodo + iFodoOff + 2, aucFpage);
			if (iSize < 6 || iBytes < iFodoOff + 8) {
				DBG_DEC(iSize);
				DBG_DEC(iFodoOff);
				iInfoLen = 2;
				break;
			}
			iCol = (int)ucGetByte(iFodo + iFodoOff + 4, aucFpage);
			if (iCol < 1 ||
			    iBytes < iFodoOff + 4 + (iCol + 1) * 2) {
				DBG_DEC(iCol);
				DBG_DEC(iFodoOff);
				iInfoLen = 2;
				break;
			}
			if (iCol >= elementsof(pRow->asColumnWidth)) {
				werr(1, "The number of columns is corrupt");
			}
			pRow->ucNumberOfColumns = iCol;
			pRow->iColumnWidthSum = 0;
			iPosPrev = (int)(short)usGetWord(
					iFodo + iFodoOff + 5,
					aucFpage);
			for (iIndex = 0; iIndex < iCol; iIndex++) {
				iPosCurr = (int)(short)usGetWord(
					iFodo + iFodoOff + 7 + iIndex * 2,
					aucFpage);
				pRow->asColumnWidth[iIndex] =
							iPosCurr - iPosPrev;
				pRow->iColumnWidthSum +=
					pRow->asColumnWidth[iIndex];
				iPosPrev = iPosCurr;
			}
			bFoundd608 = TRUE;
			break;
		default:
			break;
		}
		if (iInfoLen <= 0) {
			iInfoLen =
				iGet8InfoLength(iFodo + iFodoOff, aucFpage);
			fail(iInfoLen <= 0);
		}
		iFodoOff += iInfoLen;
	}
	if (bFound2416 && bFound2417 && bFoundd608) {
		return found_end_of_row;
	}
	if (bFound2416) {
		return found_a_cell;
	}
	return found_nothing;
} /* end of eGet8RowInfo */

/*
 * Fill the style information block with information
 * from a Word 8/97 file.
 * Returns TRUE when successful, otherwise FALSE
 */
static BOOL
bGet8StyleInfo(int iFodo, const unsigned char *aucFpage,
		style_block_type *pStyle)
{
	int	iBytes, iFodoOff, iInfoLen;

	fail(iFodo <= 0 || aucFpage == NULL || pStyle == NULL);

	iBytes = 2 * (int)ucGetByte(iFodo, aucFpage);
	iFodoOff = 3;
	if (iBytes == 0) {
		iFodo++;
		iBytes = 2 * (int)ucGetByte(iFodo, aucFpage);
	}
	NO_DBG_DEC(iBytes);
	if (iBytes < 2) {
		return FALSE;
	}
	(void)memset(pStyle, 0, sizeof(*pStyle));
	pStyle->ucStyle = ucGetByte(iFodo + 1, aucFpage);
	NO_DBG_DEC(pStyle->ucStyle);
	while (iBytes >= iFodoOff + 2) {
		iInfoLen = 0;
		switch (usGetWord(iFodo + iFodoOff, aucFpage)) {
		case 0x260a:
			pStyle->bInList = TRUE;
			pStyle->ucListLevel =
				ucGetByte(iFodo + iFodoOff + 2, aucFpage);
			DBG_DEC(pStyle->ucListLevel);
			break;
		case 0x460b:
			DBG_DEC(usGetWord(iFodo + iFodoOff + 2, aucFpage));
			break;
		case 0x840f:
			pStyle->sLeftIndent = (short)usGetWord(
					iFodo + iFodoOff + 2, aucFpage);
			NO_DBG_DEC(pStyle->sLeftIndent);
			break;
		default:
			break;
		}
		if (iInfoLen <= 0) {
			iInfoLen =
				iGet8InfoLength(iFodo + iFodoOff, aucFpage);
			fail(iInfoLen <= 0);
		}
		iFodoOff += iInfoLen;
	}
	return TRUE;
} /* end of bGet8StyleInfo */

/*
 * Build the lists with Paragraph Information for Word 8/97 files
 */
static void
vGet8ParInfo(FILE *pFile, const pps_info_type *pPPS,
	const int *aiBBD, int iBBDLen, const int *aiSBD, int iSBDLen,
	const unsigned char *aucHeader)
{
	row_block_type		tRow;
	style_block_type	tStyle;
	const int	*aiBlockDepot;
	int		*aiParfPage;
	unsigned char	*aucBuffer;
	size_t		tSize;
	int	iTableStartBlock, iTableSize, iBlockDepotLen, iBlockSize;
	int	iLen, iOffset;
	int	iParmOffset, iParmOffsetFirst, iParmOffsetLast;
	int	iIndex, iIndex2, iNbr, iFodo;
	int	iBeginParfInfo, iParfInfoLen;
	unsigned short	usDocStatus;
	unsigned char	aucFpage[BIG_BLOCK_SIZE];

	fail(pFile == NULL || pPPS == NULL || aucHeader == NULL);
	fail(aiBBD == NULL || iBBDLen < 0);
	fail(aiSBD == NULL || iSBDLen < 0);

	iBeginParfInfo = (int)ulGetLong(0x102, aucHeader);
	DBG_HEX(iBeginParfInfo);
	iParfInfoLen = (int)ulGetLong(0x106, aucHeader);
	DBG_DEC(iParfInfoLen);

	/* Use 0Table or 1Table? */
	usDocStatus = usGetWord(0x0a, aucHeader);
	if (usDocStatus & BIT(9)) {
		iTableStartBlock = pPPS->t1Table.iSb;
		iTableSize = pPPS->t1Table.iSize;
	} else {
		iTableStartBlock = pPPS->t0Table.iSb;
		iTableSize = pPPS->t0Table.iSize;
	}
	DBG_DEC(iTableStartBlock);
	if (iTableStartBlock < 0) {
		DBG_DEC(iTableStartBlock);
		DBG_MSG("No paragraph information");
		return;
	}
	DBG_HEX(iTableSize);
	if (iTableSize < MIN_SIZE_FOR_BBD_USE) {
	  	/* Use the Small Block Depot */
		aiBlockDepot = aiSBD;
		iBlockDepotLen = iSBDLen;
		iBlockSize = SMALL_BLOCK_SIZE;
	} else {
	  	/* Use the Big Block Depot */
		aiBlockDepot = aiBBD;
		iBlockDepotLen = iBBDLen;
		iBlockSize = BIG_BLOCK_SIZE;
	}
	aucBuffer = xmalloc(iParfInfoLen);
	if (!bReadBuffer(pFile, iTableStartBlock,
			aiBlockDepot, iBlockDepotLen, iBlockSize,
			aucBuffer, iBeginParfInfo, iParfInfoLen)) {
		xfree(aucBuffer);
		return;
	}
	NO_DBG_PRINT_BLOCK(aucBuffer, iParfInfoLen);

	iLen = ((iParfInfoLen / 4) - 1) / 2;
	tSize = iLen * sizeof(int);
	aiParfPage = xmalloc(tSize);
	for (iIndex = 0, iOffset = (iLen + 1) * 4;
	     iIndex < iLen;
	     iIndex++, iOffset += 4) {
		 aiParfPage[iIndex] = (int)ulGetLong(iOffset, aucBuffer);
		 NO_DBG_DEC(aiParfPage[iIndex]);
	}
	DBG_HEX(ulGetLong(0, aucBuffer));
	xfree(aucBuffer);
	aucBuffer = NULL;
	NO_DBG_PRINT_BLOCK(aucHeader, HEADER_SIZE);

	(void)memset(&tStyle, 0, sizeof(tStyle));
	(void)memset(&tRow, 0, sizeof(tRow));
	iParmOffsetFirst = -1;
	for (iIndex = 0; iIndex < iLen; iIndex++) {
		fail(aiParfPage[iIndex] > INT_MAX / BIG_BLOCK_SIZE);
		if (!bReadBuffer(pFile, pPPS->tWordDocument.iSb,
				aiBBD, iBBDLen, BIG_BLOCK_SIZE,
				aucFpage,
				aiParfPage[iIndex] * BIG_BLOCK_SIZE,
				BIG_BLOCK_SIZE)) {
			break;
		}
		NO_DBG_PRINT_BLOCK(aucFpage, BIG_BLOCK_SIZE);
		iNbr = (int)ucGetByte(0x1ff, aucFpage);
		NO_DBG_DEC(iNbr);
		for (iIndex2 = 0; iIndex2 < iNbr; iIndex2++) {
			NO_DBG_HEX(ulGetLong(iIndex2 * 4, aucFpage));
			iFodo = 2 * (int)ucGetByte(
				(iNbr + 1) * 4 + iIndex2 * 13, aucFpage);
			if (iFodo <= 0) {
				continue;
			}
			if (bGet8StyleInfo(iFodo, aucFpage, &tStyle)) {
				iParmOffset = (int)ulGetLong(
						iIndex2 * 4, aucFpage);
				NO_DBG_HEX(iParmOffset);
				tStyle.iOffset = iTextOffset2FileOffset(
							iParmOffset);
				vAdd2StyleInfoList(&tStyle);
				(void)memset(&tStyle, 0, sizeof(tStyle));
			}
			switch (eGet8RowInfo(iFodo, aucFpage, &tRow)) {
			case found_a_cell:
				if (iParmOffsetFirst >= 0) {
					break;
				}
				iParmOffsetFirst = (int)ulGetLong(
						iIndex2 * 4, aucFpage);
				NO_DBG_HEX(iParmOffsetFirst);
				tRow.iOffsetStart = iTextOffset2FileOffset(
							iParmOffsetFirst);
				NO_DBG_HEX(tRow.iOffsetStart);
				break;
			case found_end_of_row:
				iParmOffsetLast = (int)ulGetLong(
						iIndex2 * 4, aucFpage);
				NO_DBG_HEX(iParmOffsetLast);
				tRow.iOffsetEnd = iTextOffset2FileOffset(
							iParmOffsetLast);
				NO_DBG_HEX(tRow.iOffsetEnd);
				vAdd2RowInfoList(&tRow);
				(void)memset(&tRow, 0, sizeof(tRow));
				iParmOffsetFirst = -1;
				break;
			default:
				break;
			}
		}
	}
	xfree(aiParfPage);
} /* end of vGet8ParInfo */

/*
 * Fill the font information block with information
 * from a Word 8/97 file.
 * Returns TRUE when successful, otherwise FALSE
 */
static BOOL
bGet8FontInfo(int iFodo, const unsigned char *aucFpage,
		font_block_type *pFont)
{
	int	iBytes, iFodoOff, iInfoLen;
	int	iTmp;

	fail(iFodo <= 0 || aucFpage == NULL || pFont == NULL);

	iBytes = (int)ucGetByte(iFodo, aucFpage);
	iFodoOff = 1;
	if (iBytes == 0) {
		iFodo++;
		iBytes = (int)ucGetByte(iFodo, aucFpage);
	}
	NO_DBG_DEC(iBytes);
	if (iBytes < 2) {
		return FALSE;
	}
	(void)memset(pFont, 0, sizeof(*pFont));
	pFont->ucFontsize = DEFAULT_FONT_SIZE;
	while (iBytes >= iFodoOff + 2) {
		iInfoLen = 0;
		switch (usGetWord(iFodo + iFodoOff, aucFpage)) {
		case 0x0835:
			iTmp = (int)ucGetByte(
					iFodo + iFodoOff + 2, aucFpage);
			if (iTmp == 0x01 || iTmp == 0x81) {
				pFont->ucFontstyle |= FONT_BOLD;
			}
			DBG_HEX_C(iTmp != 0x01 && iTmp != 0x81, iTmp);
			break;
		case 0x0836:
			iTmp = (int)ucGetByte(
					iFodo + iFodoOff + 2, aucFpage);
			if (iTmp == 0x01 || iTmp == 0x81) {
				pFont->ucFontstyle |= FONT_ITALIC;
			}
			DBG_HEX_C(iTmp != 0x01 && iTmp != 0x81, iTmp);
			break;
		case 0x083a:
		case 0x083b:
			iTmp = (int)ucGetByte(
					iFodo + iFodoOff + 2, aucFpage);
			if (iTmp == 0x01 || iTmp == 0x81) {
				DBG_MSG("All capitals text");
				pFont->ucFontstyle |= FONT_CAPITALS;
			}
			DBG_HEX_C(iTmp != 0x01 && iTmp != 0x81, iTmp);
			break;
		case 0x083c:
			iTmp = (int)ucGetByte(
					iFodo + iFodoOff + 2, aucFpage);
			if (iTmp == 0x01 || iTmp == 0x81) {
				DBG_MSG("Hidden text");
				pFont->ucFontstyle |= FONT_HIDDEN;
			}
			DBG_HEX_C(iTmp != 0x01 && iTmp != 0x81, iTmp);
			break;
		case 0x4a43:
			iTmp = (int)usGetWord(
					iFodo + iFodoOff + 2, aucFpage);
			if (iTmp < 0) {
				pFont->ucFontsize = 0;
			} else if (iTmp > UCHAR_MAX) {
				pFont->ucFontsize = UCHAR_MAX;
			} else {
				pFont->ucFontsize = (unsigned char)iTmp;
			}
			break;
		case 0x4a51:
			iTmp = (int)usGetWord(
					iFodo + iFodoOff + 2, aucFpage);
			if (iTmp >= 0 && iTmp <= UCHAR_MAX) {
				pFont->ucFontnumber = (unsigned char)iTmp;
			}
			break;
		default:
			break;
		}
		if (iInfoLen <= 0) {
			iInfoLen =
				iGet8InfoLength(iFodo + iFodoOff, aucFpage);
			fail(iInfoLen <= 0);
		}
		iFodoOff += iInfoLen;
	}
	return TRUE;
} /* end of bGet8FontInfo */

/*
 * Build the lists with Character Information for Word 8/97 files
 */
static void
vGet8ChrInfo(FILE *pFile, const pps_info_type *pPPS,
	const int *aiBBD, int iBBDLen, const int *aiSBD, int iSBDLen,
	const unsigned char *aucHeader)
{
	font_block_type	tFont;
	const int	*aiBlockDepot;
	int		*aiCharPage;
	unsigned char	*aucBuffer;
	size_t	tSize;
	int	iTableStartBlock, iTableSize, iBlockDepotLen, iBlockSize;
	int	iLen, iOffset;
	int	iParmOffset, iFileOffset;
	int	iIndex, iIndex2, iNbr, iFodo;
	int	iBeginCharInfo, iCharInfoLen;
	unsigned short	usDocStatus;
	unsigned char	aucFpage[BIG_BLOCK_SIZE];

	fail(pFile == NULL || pPPS == NULL || aucHeader == NULL);
	fail(aiBBD == NULL || iBBDLen < 0);

	iBeginCharInfo = (int)ulGetLong(0xfa, aucHeader);
	DBG_HEX(iBeginCharInfo);
#if defined(__hpux)
	/* Strange: the HPUX cpp can not deal with the number 0xfe */
	iCharInfoLen = (int)aucHeader[0xff] << 8 | (int)aucHeader[0xfe];
#else
	iCharInfoLen = (int)ulGetLong(0xfe, aucHeader);
#endif /* __hpux */
	DBG_DEC(iCharInfoLen);

	/* Use 0Table or 1Table? */
	usDocStatus = usGetWord(0x0a, aucHeader);
	if (usDocStatus & BIT(9)) {
		iTableStartBlock = pPPS->t1Table.iSb;
		iTableSize = pPPS->t1Table.iSize;
	} else {
		iTableStartBlock = pPPS->t0Table.iSb;
		iTableSize = pPPS->t0Table.iSize;
	}
	DBG_DEC(iTableStartBlock);
	if (iTableStartBlock < 0) {
		DBG_DEC(iTableStartBlock);
		DBG_MSG("No character information");
		return;
	}
	DBG_HEX(iTableSize);
	if (iTableSize < MIN_SIZE_FOR_BBD_USE) {
	  	/* Use the Small Block Depot */
		aiBlockDepot = aiSBD;
		iBlockDepotLen = iSBDLen;
		iBlockSize = SMALL_BLOCK_SIZE;
	} else {
	  	/* Use the Big Block Depot */
		aiBlockDepot = aiBBD;
		iBlockDepotLen = iBBDLen;
		iBlockSize = BIG_BLOCK_SIZE;
	}
	aucBuffer = xmalloc(iCharInfoLen);
	if (!bReadBuffer(pFile, iTableStartBlock,
			aiBlockDepot, iBlockDepotLen, iBlockSize,
			aucBuffer, iBeginCharInfo, iCharInfoLen)) {
		xfree(aucBuffer);
		return;
	}
	NO_DBG_PRINT_BLOCK(aucBuffer, iCharInfoLen);

	iLen = ((iCharInfoLen / 4) - 1) / 2;
	tSize = iLen * sizeof(int);
	aiCharPage = xmalloc(tSize);
	for (iIndex = 0, iOffset = (iLen + 1) * 4;
	     iIndex < iLen;
	     iIndex++, iOffset += 4) {
		 aiCharPage[iIndex] = (int)ulGetLong(iOffset, aucBuffer);
		 DBG_DEC(aiCharPage[iIndex]);
	}
	DBG_HEX(ulGetLong(0, aucBuffer));
	xfree(aucBuffer);
	aucBuffer = NULL;
	NO_DBG_PRINT_BLOCK(aucHeader, HEADER_SIZE);

	(void)memset(&tFont, 0, sizeof(tFont));
	for (iIndex = 0; iIndex < iLen; iIndex++) {
		fail(aiCharPage[iIndex] > INT_MAX / BIG_BLOCK_SIZE);
		if (!bReadBuffer(pFile, pPPS->tWordDocument.iSb,
				aiBBD, iBBDLen, BIG_BLOCK_SIZE,
				aucFpage,
				aiCharPage[iIndex] * BIG_BLOCK_SIZE,
				BIG_BLOCK_SIZE)) {
			break;
		}
		NO_DBG_PRINT_BLOCK(aucFpage, BIG_BLOCK_SIZE);
		iNbr = (int)ucGetByte(0x1ff, aucFpage);
		NO_DBG_DEC(iNbr);
		for (iIndex2 = 0; iIndex2 < iNbr; iIndex2++) {
			iParmOffset = (int)ulGetLong(
						iIndex2 * 4, aucFpage);
			iFileOffset = iTextOffset2FileOffset(iParmOffset);
			iFodo = 2 * (int)ucGetByte(
				(iNbr + 1) * 4 + iIndex2, aucFpage);
			if (iFodo == 0) {
				NO_DBG_HEX(iParmOffset);
				vReset2FontInfoList(iFileOffset);
			} else {
				if (bGet8FontInfo(iFodo, aucFpage, &tFont)) {
					tFont.iOffset = iFileOffset;
					vAdd2FontInfoList(&tFont);
				}
			}
			(void)memset(&tFont, 0, sizeof(tFont));
		}
	}
	xfree(aiCharPage);
} /* end of vGet8ChrInfo */

/*
 * Build the lists with Paragraph Information
 */
void
vGetParagraphInfo(FILE *pFile, const pps_info_type *pPPS,
		const int *aiBBD, int iBBDLen, const int *aiSBD, int iSBDLen,
		const unsigned char *aucHeader, int iWordVersion)
{
	options_type	tOptions;

	fail(pFile == NULL || pPPS == NULL || aucHeader == NULL);
	fail(iWordVersion < 6 || iWordVersion > 8);
	fail(aiBBD == NULL || iBBDLen < 0);
	fail(aiSBD == NULL || iSBDLen < 0);

	vGetOptions(&tOptions);
	switch (iWordVersion) {
	case 6:
	case 7:
		vGet6ParInfo(pFile, pPPS->tWordDocument.iSb,
			aiBBD, iBBDLen, aucHeader);
		if (tOptions.bUseOutlineFonts) {
			vGet6ChrInfo(pFile, pPPS->tWordDocument.iSb,
				aiBBD, iBBDLen, aucHeader);
#if defined(__riscos)
			vCreate6FontTable(pFile, pPPS->tWordDocument.iSb,
				aiBBD, iBBDLen, aucHeader);
#endif /* __riscos */
		}
		break;
	case 8:
		vGet8ParInfo(pFile, pPPS,
			aiBBD, iBBDLen, aiSBD, iSBDLen, aucHeader);
		if (tOptions.bUseOutlineFonts) {
			vGet8ChrInfo(pFile, pPPS,
				aiBBD, iBBDLen, aiSBD, iSBDLen, aucHeader);
#if defined(__riscos)
			vCreate8FontTable(pFile, pPPS,
				aiBBD, iBBDLen, aiSBD, iSBDLen, aucHeader);
#endif /* __riscos */
		}
		break;
	default:
		werr(0, "Sorry, no paragraph information");
		break;
	}
} /* end of vGetParagraphInfo */
