/*
 * wordlib.c
 * Copyright (C) 1998,1999 A.J. van Os
 *
 * Description:
 * Deal with the internals of a MS Word file
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "antiword.h"

static BOOL	bMacFile = FALSE;

/* Macro to make sure all such statements will be identical */
#define FREE_ALL()		\
	do {\
		vDestroySmallBlockList();\
		xfree(aiRootList);\
		aiRootList = NULL;\
		xfree(aiSbdList);\
		aiSbdList = NULL;\
		xfree(aiBbdList);\
		aiBbdList = NULL;\
		xfree(aiSBD);\
		aiSBD = NULL;\
		xfree(aiBBD);\
		aiBBD = NULL;\
	} while(0)


/*
 * ulReadLong - read four bytes from the given file and offset
 */
static unsigned long
ulReadLong(FILE *pFile, long lOffset)
{
	unsigned char	aucBytes[4];

	fail(pFile == NULL || lOffset < 0);

	if (!bReadBytes(aucBytes, 4, lOffset, pFile)) {
		werr(1, "Read long %ld not possible", lOffset);
	}
	return ulGetLong(0, aucBytes);
} /* end of ulReadLong */

/*
 * vName2String - turn the name into a proper string.
 */
static void
vName2String(int iNameSize, const unsigned char *aucBytes, char *szName)
{
	char	*pcChar;
	int	iIndex;

	fail(aucBytes == NULL || szName == NULL);

	if (iNameSize <= 0) {
		szName[0] = '\0';
		return;
	}
	for (iIndex = 0, pcChar = szName;
	     iIndex < 2 * iNameSize;
	     iIndex += 2, pcChar++) {
		*pcChar = (char)aucBytes[iIndex];
	}
	szName[iNameSize - 1] = '\0';
} /* end of vName2String */

/*
 * iReadBlockIndices - read the Big/Small Block Depot indices
 *
 * returns the number of indices read
 */
static int
iReadBlockIndices(FILE *pFile, int *aiBlockDepot, int iMaxRec, int iOffset)
{
	int	iIndex, iDone;
	unsigned char	aucBytes[BIG_BLOCK_SIZE];

	fail(pFile == NULL || aiBlockDepot == NULL);
	fail(iMaxRec < 0 || iOffset < 0);

	/* Read a big block with BBD or SBD indices */
	if (!bReadBytes(aucBytes, BIG_BLOCK_SIZE, iOffset, pFile)) {
		werr(0, "Read big block %d is not possible", iOffset);
		return -1;
	}
	/* Split the big block into indices, an index is four bytes */
	iDone = min(iMaxRec, BIG_BLOCK_SIZE / 4);
	for (iIndex = 0; iIndex < iDone; iIndex++) {
		aiBlockDepot[iIndex] = (int)ulGetLong(4 * iIndex, aucBytes);
		NO_DBG_DEC(aiBlockDepot[iIndex]);
	}
	return iDone;
} /* end of iReadBlockIndices */

/*
 * bGetBBD - get the Big Block Depot indices from the index-blocks
 */
static BOOL
bGetBBD(FILE *pFile, const int *aiDepot, int iDepotLen,
			int *aiBBD, int iBBDLen)
{
	int	iIndex, iBegin, iToGo, iDone;

	fail(pFile == NULL || aiDepot == NULL || aiBBD == NULL);
	fail(iDepotLen < 0 || iBBDLen < 0);

	DBG_MSG("bGetBBD");

	iToGo = iBBDLen;
	for (iIndex = 0; iIndex < iDepotLen && iToGo > 0; iIndex++) {
		iBegin = (aiDepot[iIndex] + 1) * BIG_BLOCK_SIZE;
		NO_DBG_HEX(iBegin);
		iDone = iReadBlockIndices(pFile, aiBBD, iToGo, iBegin);
		if (iDone < 0) {
			return FALSE;
		}
		aiBBD += iDone;
		iToGo -= iDone;
	}
	return iToGo <= 0;
} /* end of bGetBBD */

/*
 * bGetSBD - get the Small Block Depot indices from the index-blocks
 */
static BOOL
bGetSBD(FILE *pFile, const int *aiDepot, int iDepotLen,
			int *aiSBD, int iSBDLen)
{
	int	iIndex, iBegin, iToGo, iDone;

	fail(pFile == NULL || aiDepot == NULL || aiSBD == NULL);
	fail(iDepotLen < 0 || iSBDLen < 0);

	DBG_MSG("bGetSBD");

	iToGo = iSBDLen;
	for (iIndex = 0; iIndex < iDepotLen && iToGo > 0; iIndex++) {
		iBegin = (aiDepot[iIndex] + 1) * BIG_BLOCK_SIZE;
		NO_DBG_HEX(iBegin);
		iDone = iReadBlockIndices(pFile, aiSBD, iToGo, iBegin);
		if (iDone < 0) {
			return FALSE;
		}
		aiSBD += iDone;
		iToGo -= iDone;
	}
	return iToGo <= 0;
} /* end of bGetSBD */

/*
 * bGetPPS - search for the Property Set Storage for three sets
 *
 * Return TRUE if the WordDocument PPS is found
 */
static BOOL
bGetPPS(FILE *pFile,
		const int *aiRootList, int iRootListLen,
		pps_info_type *pPPS)
{
	pps_type	pps;
	int	iTmp, iOffset, iStartBlock, iBegin, iType, iNameSize;
	BOOL	bResult, bExcel;
	char	szName[20];
	unsigned char	aucBytes[PROPERTY_SET_STORAGE_SIZE];

	fail(pFile == NULL || pPPS == NULL);
	fail(aiRootList == NULL || iRootListLen < 0);

	DBG_MSG("bGetPPS");
	NO_DBG_DEC(iRootListLen);

	bResult = FALSE;
	bExcel = FALSE;
	(void)memset(pPPS, 0, sizeof(*pPPS));
	pPPS->tWordDocument.iSb = -1;
	pPPS->t0Table.iSb = -1;
	pPPS->t1Table.iSb = -1;

	for (iTmp = 0;
	     iTmp < iRootListLen * BIG_BLOCK_SIZE;
	     iTmp += PROPERTY_SET_STORAGE_SIZE) {
		iStartBlock = iTmp / BIG_BLOCK_SIZE;
		iOffset = iTmp % BIG_BLOCK_SIZE;
		iBegin = (aiRootList[iStartBlock] + 1) * BIG_BLOCK_SIZE +
			iOffset;
		NO_DBG_HEX(iBegin);
		if (!bReadBytes(aucBytes, PROPERTY_SET_STORAGE_SIZE,
							iBegin, pFile)) {
			werr(0, "Read ppss %d is not possible", iBegin);
			return FALSE;
		}
		iType = (int)ucGetByte(0x42, aucBytes);
		NO_DBG_DEC(iType);
		if (iType != 2) {
			continue;
		}
		iNameSize = (int)usGetWord(0x40, aucBytes);
		NO_DBG_DEC(iNameSize);
		iNameSize = (iNameSize + 1) / 2;
		iNameSize = min(iNameSize, sizeof(szName));
		NO_DBG_DEC(iNameSize);
		vName2String(iNameSize, aucBytes, szName);
		NO_DBG_MSG(szName);
		pps.iSb = (int)ulGetLong(0x74, aucBytes);
		NO_DBG_DEC(pps.iSb);
		pps.iSize = (int)ulGetLong(0x78, aucBytes);
		NO_DBG_DEC(pps.iSize);
		if (pPPS->tWordDocument.iSb < 0 &&
		    StrEq(szName, "WordDocument")) {
			pPPS->tWordDocument = pps;
			bResult = TRUE;
		} else if (pPPS->t0Table.iSb < 0 &&
			   StrEq(szName, "0Table")) {
			pPPS->t0Table = pps;
			DBG_DEC(pPPS->t0Table.iSb);
			DBG_HEX(pPPS->t0Table.iSize);
		} else if (pPPS->t1Table.iSb < 0 &&
			   StrEq(szName, "1Table")) {
			pPPS->t1Table = pps;
			DBG_DEC(pPPS->t1Table.iSb);
			DBG_HEX(pPPS->t1Table.iSize);
		} else if (StrEq(szName, "Book") ||
			   StrEq(szName, "Workbook")) {
			bExcel = TRUE;
		}
	}
	if (bResult) {
		return TRUE;
	}
	if (bExcel) {
		werr(0, "Sorry, but this is an Excel spreadsheet");
	} else {
		werr(0, "This OLE file does not contain a Word document");
	}
	return FALSE;
} /* end of bGetPPS */

/*
 * bGetDocumentText - make a list of the text blocks of a Word document
 */
static BOOL
bGetDocumentText(FILE *pFile, const pps_info_type *pPPS,
		const int *aiBBD, int iBBDLen,
		const int *aiSBD, int iSBDLen,
		const unsigned char *aucHeader, int iWordVersion)
{
	int	iBeginOfText;
	int	iTextLen, iFootnoteLen, iEndnoteLen;
	int	iHeaderLen, iMacroLen, iAnnotationLen;
	int	iTextBoxLen, iHdrTextBoxLen;
	BOOL	bFastSaved, bEncrypted, bSuccess;
	text_info_enum	eResult;
	unsigned short	usDocStatus;

	fail(pFile == NULL || pPPS == NULL);
	fail(aiBBD == NULL || iBBDLen < 0);

	DBG_MSG("bGetDocumentText");

	/* Get the status flags from the header */
	usDocStatus = usGetWord(0x0a, aucHeader);
	DBG_HEX(usDocStatus);
	bFastSaved = (usDocStatus & BIT(2)) != 0;
	DBG_MSG_C(bFastSaved, "This document is Fast Saved");
	DBG_DEC_C(bFastSaved, (usDocStatus & 0xf0) >> 4);
	bEncrypted = (usDocStatus & BIT(8)) != 0;
	if (bEncrypted) {
		werr(0, "Encrypted documents are not supported");
		return FALSE;
	}

	/* Get length information */
	iBeginOfText = (int)ulGetLong(0x18, aucHeader);
	DBG_HEX(iBeginOfText);
	if (iWordVersion == 6 || iWordVersion == 7) {
		iTextLen = (int)ulGetLong(0x34, aucHeader);
		iFootnoteLen = (int)ulGetLong(0x38, aucHeader);
		iHeaderLen = (int)ulGetLong(0x3c, aucHeader);
		iMacroLen = (int)ulGetLong(0x40, aucHeader);
		iAnnotationLen = (int)ulGetLong(0x44, aucHeader);
		iEndnoteLen = (int)ulGetLong(0x48, aucHeader);
		iTextBoxLen = (int)ulGetLong(0x4c, aucHeader);
		iHdrTextBoxLen = (int)ulGetLong(0x50, aucHeader);
	} else {
		iTextLen = (int)ulGetLong(0x4c, aucHeader);
		iFootnoteLen = (int)ulGetLong(0x50, aucHeader);
		iHeaderLen = (int)ulGetLong(0x54, aucHeader);
		iMacroLen = (int)ulGetLong(0x58, aucHeader);
		iAnnotationLen = (int)ulGetLong(0x5c, aucHeader);
		iEndnoteLen = (int)ulGetLong(0x60, aucHeader);
		iTextBoxLen = (int)ulGetLong(0x64, aucHeader);
		iHdrTextBoxLen = (int)ulGetLong(0x68, aucHeader);
	}
	DBG_DEC(iTextLen);
	DBG_DEC(iFootnoteLen);
	DBG_DEC(iHeaderLen);
	DBG_DEC(iMacroLen);
	DBG_DEC(iAnnotationLen);
	DBG_DEC(iEndnoteLen);
	DBG_DEC(iTextBoxLen);
	DBG_DEC(iHdrTextBoxLen);

	/* Make a list of the text blocks */
	switch (iWordVersion) {
	case 6:
	case 7:
		if (bFastSaved) {
			eResult = eGet6DocumentText(pFile,
					pPPS->tWordDocument.iSb,
					aiBBD, iBBDLen,
					aucHeader);
			bSuccess = eResult == text_success;
		} else {
		  	bSuccess = bAddBlocks(iBeginOfText,
				iTextLen +
				iFootnoteLen +
				iHeaderLen + iMacroLen + iAnnotationLen +
				iEndnoteLen +
				iTextBoxLen + iHdrTextBoxLen,
				FALSE,
				pPPS->tWordDocument.iSb,
				aiBBD, iBBDLen);
		}
		break;
	case 8:
		eResult = eGet8DocumentText(pFile,
				pPPS,
				aiBBD, iBBDLen, aiSBD, iSBDLen,
				aucHeader);
		if (eResult == text_no_information && !bFastSaved) {
			/* Assume there is only one block */
			bSuccess = bAddBlocks(iBeginOfText,
				iTextLen +
				iFootnoteLen +
				iHeaderLen + iMacroLen + iAnnotationLen +
				iEndnoteLen +
				iTextBoxLen + iHdrTextBoxLen,
				FALSE,
				pPPS->tWordDocument.iSb,
				aiBBD, iBBDLen);
		} else {
			bSuccess = eResult == text_success;
		}
		break;
	default:
		werr(0, "This version of Word is not supported");
		bSuccess = FALSE;
		break;
	}

	if (bSuccess) {
		vSplitBlockList(iTextLen,
				iFootnoteLen,
				iHeaderLen + iMacroLen + iAnnotationLen,
				iEndnoteLen,
				iTextBoxLen + iHdrTextBoxLen,
				!bFastSaved && iWordVersion == 8);
	} else {
		vDestroyTextBlockList();
		werr(0, "I can't find the text of this document");
	}
	return bSuccess;
} /* end of bGetDocumentText */

/*
 * pInitFile - open and check the file
 */
static FILE *
pInitFile(const char *szFilename)
{
	FILE	*pFile;

	fail(szFilename == NULL || szFilename[0] == '\0');

	pFile = fopen(szFilename, "rb");
	if (pFile == NULL) {
		werr(0, "I can't open '%s' for reading", szFilename);
		return NULL;
	}
	return pFile;
} /* end of pInitFile */

static BOOL
bInitDoc(FILE *pFile, off_t tFilesize)
{
	pps_info_type	PPS_info;
	int	*aiBBD, *aiRootList, *aiBbdList;
	int	*aiSBD, *aiSbdList;
	size_t	tSize;
	int	iRootListLen, iRootStartblock;
	int	iSbdStartblock, iSBLstartblock;
	int	iBBDLen, iSBDLen;
	int	iWordVersion, iIndex, iTmp;
	int	iMaxBlock, iNumBbdBlocks, iMaxSmallBlock;
	BOOL	bResult;
	unsigned short	usIdent, usFib;
	unsigned char	aucHeader[HEADER_SIZE];

	fail(pFile == NULL);

	iMaxBlock = (int)(tFilesize / BIG_BLOCK_SIZE) - 2;
	DBG_DEC(iMaxBlock);
	if (iMaxBlock < 1) {
		return FALSE;
	}
	iBBDLen = iMaxBlock + 1;
	iNumBbdBlocks = (int)ulReadLong(pFile, 0x2c);
	DBG_DEC(iNumBbdBlocks);
	if (iNumBbdBlocks > 109) {
		DBG_DEC(tFilesize);
		if (tFilesize > 109 * (BIG_BLOCK_SIZE/4) * BIG_BLOCK_SIZE) {
			werr(0, "I'm afraid this file is too big to handle.");
		} else {
			werr(0, "This file is probably corrupt. Sorry.");
		}
		return FALSE;
	}
	iRootStartblock = (int)ulReadLong(pFile, 0x30);
	DBG_DEC(iRootStartblock);
	iSbdStartblock = (int)ulReadLong(pFile, 0x3c);
	DBG_DEC(iSbdStartblock);
	iSBLstartblock = (int)ulReadLong(pFile,
		((long)iRootStartblock + 1) * BIG_BLOCK_SIZE + 0x74);
	DBG_DEC(iSBLstartblock);
	iMaxSmallBlock = (int)(ulReadLong(pFile,
		((long)iRootStartblock + 1) *
		BIG_BLOCK_SIZE + 0x78) / SMALL_BLOCK_SIZE) - 1;
	DBG_DEC(iMaxSmallBlock);
	iSBDLen = iMaxSmallBlock + 1;
	/* All to be xmalloc-ed pointers to NULL */
	aiRootList = NULL;
	aiSbdList = NULL;
	aiBbdList = NULL;
	aiSBD = NULL;
	aiBBD = NULL;
/* Big Block Depot */
	tSize = iNumBbdBlocks * sizeof(int);
	aiBbdList = xmalloc(tSize);
	tSize = iBBDLen * sizeof(int);
	aiBBD = xmalloc(tSize);
	for (iIndex = 0; iIndex < iNumBbdBlocks; iIndex++) {
		aiBbdList[iIndex] =
			(int)ulReadLong(pFile, 0x4c + 4 * (long)iIndex);
		DBG_HEX(aiBbdList[iIndex]);
	}
	if (!bGetBBD(pFile, aiBbdList, iNumBbdBlocks, aiBBD, iBBDLen)) {
		FREE_ALL();
		return FALSE;
	}
	xfree(aiBbdList);
	aiBbdList = NULL;
/* Small Block Depot */
	tSize = iBBDLen * sizeof(int);
	aiSbdList = xmalloc(tSize);
	tSize = iSBDLen * sizeof(int);
	aiSBD = xmalloc(tSize);
	for (iIndex = 0, iTmp = iSbdStartblock;
	     iIndex < iBBDLen && iTmp != END_OF_CHAIN;
	     iIndex++, iTmp = aiBBD[iTmp]) {
		if (iTmp < 0 || iTmp >= iBBDLen) {
			werr(1, "The Big Block Depot is corrupt");
		}
		aiSbdList[iIndex] = iTmp;
		DBG_HEX(aiSbdList[iIndex]);
	}
	if (!bGetSBD(pFile, aiSbdList, iBBDLen, aiSBD, iSBDLen)) {
		FREE_ALL();
		return FALSE;
	}
	xfree(aiSbdList);
	aiSbdList = NULL;
/* Root list */
	for (iRootListLen = 0, iTmp = iRootStartblock;
	     iRootListLen < iBBDLen && iTmp != END_OF_CHAIN;
	     iRootListLen++, iTmp = aiBBD[iTmp]) {
		if (iTmp < 0 || iTmp >= iBBDLen) {
			werr(1, "The Big Block Depot is corrupt");
		}
	}
	if (iRootListLen <= 0) {
		werr(0, "No Rootlist found");
		FREE_ALL();
		return FALSE;
	}
	tSize = iRootListLen * sizeof(int);
	aiRootList = xmalloc(tSize);
	for (iIndex = 0, iTmp = iRootStartblock;
	     iIndex < iBBDLen && iTmp != END_OF_CHAIN;
	     iIndex++, iTmp = aiBBD[iTmp]) {
		if (iTmp < 0 || iTmp >= iBBDLen) {
			werr(1, "The Big Block Depot is corrupt");
		}
		aiRootList[iIndex] = iTmp;
		NO_DBG_DEC(aiRootList[iIndex]);
	}
	fail(iRootListLen != iIndex);
	bResult = bGetPPS(pFile, aiRootList, iRootListLen, &PPS_info);
	xfree(aiRootList);
	aiRootList = NULL;
	if (!bResult) {
		FREE_ALL();
		return FALSE;
	}
/* Small block list */
	if (!bCreateSmallBlockList(iSBLstartblock, aiBBD, iBBDLen)) {
		FREE_ALL();
		return FALSE;
	}

	if (PPS_info.tWordDocument.iSize < MIN_SIZE_FOR_BBD_USE) {
		DBG_DEC(PPS_info.tWordDocument.iSize);
		FREE_ALL();
		werr(0, "I'm afraid the text stream of this file "
			"is too small to handle.");
		return FALSE;
	}
	/* Read the headerblock */
	if (!bReadBuffer(pFile, PPS_info.tWordDocument.iSb,
			aiBBD, iBBDLen, BIG_BLOCK_SIZE,
			aucHeader, 0, HEADER_SIZE)) {
		FREE_ALL();
		return FALSE;
	}
	usIdent = usGetWord(0x00, aucHeader);
	DBG_HEX(usIdent);
	fail(usIdent != 0xa5dc &&	/* Word 6 & 7; Windows */
	     usIdent != 0xa5ec &&	/* */
	     usIdent != 0xa697);	/* Word 6; Mac; FastSaved */
	usFib = usGetWord(0x02, aucHeader);
	DBG_DEC(usFib);
	if (usFib < 101) {
		FREE_ALL();
		werr(0, "This file is from a version of Word before Word 6.");
		return FALSE;
	}
	bMacFile = FALSE;
	/* Get the version of Word */
	switch (usFib) {
	case 101:
	case 102:
		iWordVersion = 6;
		DBG_MSG("Word 6 for Windows");
		break;
	case 103:
	case 104:
		if (ucGetByte(0x05, aucHeader) == 0xe0) {
			iWordVersion = 7;
			DBG_MSG("Word 7 for Win95");
		} else {
			iWordVersion = 6;
			DBG_MSG("Word 6 for Macintosh");
			bMacFile = TRUE;
		}
		break;
	default:
		iWordVersion = 8;
		DBG_MSG("Word97 for Win95/98/NT");
		break;
	}
	bResult = bGetDocumentText(pFile, &PPS_info,
			aiBBD, iBBDLen, aiSBD, iSBDLen,
			aucHeader, iWordVersion);
	if (bResult) {
		vGetParagraphInfo(pFile, &PPS_info,
			aiBBD, iBBDLen, aiSBD, iSBDLen,
			aucHeader, iWordVersion);
		vGetNotesInfo(pFile, &PPS_info,
			aiBBD, iBBDLen, aiSBD, iSBDLen,
			aucHeader, iWordVersion);
	}
	FREE_ALL();
	return bResult;
} /* end of bInitDoc */

/*
 * pOpenDocument - open and initialize the document
 */
FILE *
pOpenDocument(const char *szFilename)
{
	FILE	*pFile;
	off_t	tFilesize;

	tFilesize = tGetFilesize(szFilename);
	if (tFilesize == 0 || tFilesize == (off_t)-1) {
		return NULL;
	}
	pFile = pInitFile(szFilename);
	if (pFile == NULL) {
		return NULL;
	}
	if (bInitDoc(pFile, tFilesize)) {
		return pFile;
	}
	(void)fclose(pFile);
	return NULL;
} /* end of pOpenDocument*/

/*
 * vCloseDocument
 */
void
vCloseDocument(FILE *pFile)
{
	DBG_MSG("vCloseDocument");

	/* Free the memory */
	vDestroyTextBlockList();
	vDestroyRowInfoList();
	vDestroyStyleInfoList();
	vDestroyFontInfoList();
	vDestroyNotesInfoLists();
#if defined(__riscos)
	vDestroyFontTable();
#endif /* __riscos */
	/* Close the file */
	(void)fclose(pFile);
} /* end of vCloseDocument */

/*
 * This function checks whether the given file is or is not a Word Document
 */
BOOL
bIsWordFile(const char *szFilename)
{
	static unsigned char	aucStartBytes[] =
		{ 0xd0, 0xcf, 0x11, 0xe0, 0xa1, 0xb1, 0x1a, 0xe1 };
	FILE	*pFile;
	off_t	tSize;
	int	iIndex, iChar;
	BOOL	bResult;

	if (szFilename == NULL || szFilename[0] == '\0') {
		return FALSE;
	}
	tSize = tGetFilesize(szFilename);
	if (tSize < 3 * BIG_BLOCK_SIZE || tSize % BIG_BLOCK_SIZE != 0) {
		return FALSE;
	}
	pFile = fopen(szFilename, "rb");
	if (pFile == NULL) {
		return FALSE;
	}
	bResult = TRUE;
	for (iIndex = 0; iIndex < elementsof(aucStartBytes); iIndex++) {
		iChar = getc(pFile);
		if (iChar == EOF || iChar != (int)aucStartBytes[iIndex]) {
			bResult = FALSE;
			break;
		}
	}
	(void)fclose(pFile);
	return bResult;
} /* end of bIsWordFile */

/*
 * This function checks whether the given file is or is not a RTF Document
 */
BOOL
bIsRtfFile(const char *szFilename)
{
	static unsigned char	aucStartBytes[] =
		{ '{', '\\', 'r', 't', 'f', '1' };
	FILE	*pFile;
	int	iIndex, iChar;
	BOOL	bResult;

	if (szFilename == NULL || szFilename[0] == '\0') {
		return FALSE;
	}
	pFile = fopen(szFilename, "rb");
	if (pFile == NULL) {
		return FALSE;
	}
	bResult = TRUE;
	for (iIndex = 0; iIndex < elementsof(aucStartBytes); iIndex++) {
		iChar = getc(pFile);
		if (iChar == EOF || iChar != (int)aucStartBytes[iIndex]) {
			bResult = FALSE;
			break;
		}
	}
	(void)fclose(pFile);
	return bResult;
} /* end of bIsRtfFile */

/*
 * TRUE is the current file was last saved on an Apple Macintosh, FALSE
 * if it was not. This function hides the methode of how to find out from
 * the rest of the program.
 *
 * note: this should grow into a real function
 */
BOOL
bFileFromTheMac(void)
{
	return bMacFile;
} /* end of bOriginatesOnMac */
